from time import sleep
import rethinkdb
from .environment import check_stage_env
from rethinkdb.net import DefaultConnection
from time import time


BRAIN_DB = "Brain"
AUDIT_DB = "Audit"
PLUGINDB = "Plugins"
JOBS = "Jobs"
TARGETS = "Targets"
OUTPUTS = "Outputs"
SELF_TEST = {BRAIN_DB: [JOBS, TARGETS, OUTPUTS],
             #AUDIT_DB: [], #Audit DB soon into latest
             PLUGINDB: []}

DEFAULT_HOSTS = {"PROD": "rethinkdb",
                 "QA": "rethinkdb",
                 "DEV": "localhost",
                 "TESTING": "localhost",
                 "": "localhost", #environment not configured, try anyway
                }


class BrainNotReady(Exception):
    """
    Simple exception to identify that the brain is not ready for use at this time
    """
    pass



def brain_post(connection, requirements=SELF_TEST):
    """
    Power On Self Test for the brain.

    Checks that the brain is appropriately seeded and ready for use.

    Raises AssertionError's if the brain is not ready.

    :param connection:  <rethinkdb.net.DefaultConnection>
    :param requirements:<dict> keys=Required Databases, key-values=Required Tables in each database
    :return: <None>
    """
    remote_dbs = [x for x in rethinkdb.db_list().run(connection)]
    for database in requirements:
        assert (database in remote_dbs), "database {} must exist".format(database)
        remote_tables = frozenset([x for x in rethinkdb.db(database).table_list().run(connection)])
        for table in requirements[database]:
            assert (table in remote_tables), "{} must exist in {}".format(table, database)

def connect(host=None,
            port=rethinkdb.DEFAULT_PORT,
            timeout=20,
            verify=True,
            **kwargs):
    """
    RethinkDB semantic connection wrapper

    raises <brain.connection.BrainNotReady> if connection verification fails

    :param verify: <bool> (default True) whether to run POST
    :param timeout: <int> max time (s) to wait for connection
    :param kwargs:  <dict> passthrough rethinkdb arguments
    :return:
    """
    if not host:
        host = DEFAULT_HOSTS.get(check_stage_env())
    connection = None
    tries = 0
    time_quit = time() + timeout
    while not connection and time() <= time_quit:
        tries += 1
        try:
            connection = rethinkdb.connect(host,
                                           port,
                                           timeout=timeout/3,
                                           **kwargs)
            if verify:
                brain_post(connection)
        except (rethinkdb.errors.ReqlDriverError, AssertionError):
            connection = None
        if not connection:
            sleep(0.5)
    if not connection:
        raise BrainNotReady("BrainNotReady: Tried {} times at {} second max timeout".format(tries, timeout))
    return connection




from os import environ

#https://en.wikipedia.org/wiki/Syslog#Severity_levels
LOGLEVELS = {"EMERGENCY": 0,
             "ALERT": 1,
             "CRITICAL": 2,
             "ERROR": 3,
             "WARNING": 4,
             "NOTICE": 5,
             "INFORMATIONAL": 6,
             "DEBUG": 7,
             "TEST": 8}

STAGES = {"PROD": 0,
          "QA": 1,
          "DEV": 2,
          "TESTING": 3}

def check_log_env():
    """
    Special environment variable LOGLEVEL may be one of 9 keys in
    <brain.environment.LOGLEVELS>

    :return: <str> (defaults to 'TEST' / most verbose)
    """
    return environ.get("LOGLEVEL", "TEST")

def check_stage_env():
    """
    Special environment variable STAGE may be one of 4 keys in
    <brain.environment.STAGES>

    :return: <str> ( defaults to 'TESTING' )
    """
    return environ.get("STAGE", "TESTING")

def log_env_gte(desired):
    """
    Boolean check if the current environment LOGLEVEL is
    at least as verbose as a desired LOGLEVEL

    :param desired: <str> one of 9 keys in <brain.enviornment.stage>
    :return: <bool>
    """
    return LOGLEVELS.get(check_log_env()) >= LOGLEVELS.get(desired, LOGLEVELS['TEST'])

def check_prod_env():
    """
    Boolean check if the environemnt is production

    :return: <bool>
    """
    return True if check_stage_env() == "PROD" else False

def check_dev_env():
    """
    Boolean check if the environment is anything other than production

    :return: <bool>
    """
    return not check_prod_env()



"""
assortment of wrapped queries
"""
from .brain_pb2 import Job, Jobs, Target, Commands
from .checks import verify
from .connection import rethinkdb as r
from .connection import connect
from decorator import decorator

RBT = r.db("Brain").table("Targets")
RBJ = r.db("Brain").table("Jobs")
RBO = r.db("Brain").table("Outputs")
RPX = r.db("Plugins")

@decorator
def wrap_rethink_errors(f, *args, **kwargs):
    try:
        return f(*args, **kwargs)
    except (r.errors.ReqlOpFailedError,
            r.errors.ReqlError) as e:
        raise ValueError(str(e))

@decorator
def wrap_connection(f, *args, **kwargs):
    """
    conn (connection) must be the last positional argument
    in all wrapped functions

    :param f: <function> to call
    :param args:  <tuple> positional arguments
    :param kwargs: <dict> keyword arguments
    :return:
    """
    if not args[-1]:
        new_args = list(args)
        new_args[-1] = connect()
        args = tuple(new_args)
    return f(*args, **kwargs)

@wrap_connection
def get_targets(conn=None):
    """
    get_brain_targets function from Brain.Targets table.

    :return: <generator> yields dict objects
    """
    targets = RBT
    results = targets.run(conn)
    for item in results:
        yield item

@wrap_connection
def get_targets_by_plugin(plugin_name, conn=None):
    """
    get_targets_by_plugin function from Brain.Targets table

    :return: <generator> yields dict objects
    """
    targets = RBT
    results = targets.filter({"PluginName": plugin_name}).run(conn)
    for item in results:
        yield item

@wrap_connection
def get_plugin_commands(plugin_name, conn=None):
    """
    get_specific_commands function queries Plugins.<PluginName> table
    the plugin name will be based off the user selection from the ui.
    It will return the query onto w2 as a dictionary nested in a list.

    :param plugin_name: <str> user's plugin selection
    :return: <generator> yields dictionaries
    """
    results = RPX.table(plugin_name).run(conn)
    for item in results:
        yield item

@wrap_connection
def get_plugin_command(plugin_name, command_name, conn=None):
    """
    get_specific_command function queries a specific CommandName

    :param plugin_name: <str> PluginName
    :param command_name: <str> CommandName
    :return: <dict>
    """
    commands = RPX.table(plugin_name).filter(
        {"CommandName": command_name}).run(conn)
    for command in commands:
        continue  # exhausting the cursor
    return command

@wrap_connection
def is_job_done(job_id, conn=None):
    """
    is_job_done function checks to if Brain.Jobs Status is 'Done'

    :param job_id: <str> id for the job
    :param conn: (optional)<connection> to run on
    :return: <dict> if job is done <false> if
    """
    result = False
    for item in RBJ.filter({'id': job_id,
                            'Status': "Done"}).run(conn):
        result = item
    return result

@wrap_connection
def get_output_content(job_id, max_size=1024, conn=None):
    """
    returns the content buffer for a job_id if that job output exists

    :param job_id: <str> id for the job
    :param max_size: <int> truncate after [max_size] bytes
    :param conn: (optional)<connection> to run on
    :return: <str> or <bytes>
    """
    content = None
    check_status = RBO.filter({"OutputJob": {'id': job_id}}).run(conn)
    for status_item in check_status:
        if max_size and "Content" in status_item and len(status_item['Content']) > max_size:
            content = "{}\n[truncated]".format(status_item['Content'][:max_size])
        elif "Content" in status_item:
            content = status_item['Content']
        else:
            content = ""
    return content

@wrap_connection
def insert_new_target(plugin_name, location_num,
                      port_num=0, optional="",
                      verify_target=True, conn=None):
    """
    insert_new_target function gets called when the user's input is validated
    and inserts a new target to Brain.Targets table.
    :param plugin_name: <str> user input plugin name
    :param location_num: <str> user input location number
    :param port_num: <str> user input port number
    :param optional: <str> user input optional
    :return: <dict> rethinkdb insert response value
    """
    target = {"PluginName": str(plugin_name),
              "Location": str(location_num),
              "Port": str(port_num),
              "Optional": {"init": str(optional)}}
    if verify_target and not verify(target, Target()):
        raise ValueError("Invalid Target")
    output = RBT.insert([target]).run(conn)
    return output

@wrap_connection
def insert_jobs(jobs, verify_jobs=True, conn=None):
    """
    insert_jobs function inserts data into Brain.Jobs table

    jobs must be in Job format

    :param jobs: <list> of Jobs
    :return: <dict> rethinkdb insert response value
    """
    assert isinstance(jobs, list)
    if verify_jobs and not verify({"Jobs": jobs}, Jobs()):
        raise ValueError("Invalid Jobs")
    inserted = RBJ.insert(jobs).run(conn)
    return inserted

@wrap_connection
def plugin_exists(plugin_name, conn=None):
    """

    :param plugin_name:
    :param conn:
    :return: <bool> whether plugin exists
    """
    return plugin_name in RPX.table_list().run(conn)

@wrap_connection
def create_plugin(plugin_name, conn=None):
    """
    Creates a new plugin

    :param plugin_name: <str>
    :param conn:
    :return: <bool> successfully inserted
    """
    if not plugin_exists(plugin_name, conn=conn):
        RPX.table_create(plugin_name,
                         primary_key="CommandName"
                         ).run(conn)
    return True

@wrap_connection
def advertise_plugin_commands(plugin_name, commands,
                              verify_commands=False,
                              conn=None):
    """

    :param plugin_name:
    :param commands:
    :param verify_commands:
    :param conn:
    :return:
    """
    assert isinstance(commands, list)
    if verify_commands and not verify({"Commands": commands},
                                      Commands()):
        raise ValueError("Invalid Commands")
    RPX.table(plugin_name).insert(commands,
                                  conflict="update"
                                  ).run(conn)

@wrap_connection
def get_next_job(plugin_name,
                 verify_job=False, conn=None):
    """

    :param plugin_name: <str>
    :param verify_job: <bool>
    :param conn: <connection> or <NoneType>
    :return: <generator> yields <dict>
    """
    job_cur = RBJ.filter((r.row["JobTarget"]["PluginName"] == plugin_name) &
                         (r.row["Status"] == "Ready")).run(conn)
    for job in job_cur:
        if verify_job and not verify(job, Job()):
            continue #to the next job... warn?
        yield job



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: brain.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='brain.proto',
  package='brain',
  serialized_pb=_b('\n\x0b\x62rain.proto\x12\x05\x62rain\"\xe4\x01\n\x07\x43ommand\x12\n\n\x02id\x18\x01 \x01(\t\x12\x13\n\x0b\x43ommandName\x18\x02 \x02(\t\x12\x0f\n\x07Tooltip\x18\x03 \x02(\t\x12\x0e\n\x06Output\x18\x04 \x02(\x08\x12$\n\x06Inputs\x18\x05 \x03(\x0b\x32\x14.brain.Command.Input\x12,\n\x0eOptionalInputs\x18\x06 \x03(\x0b\x32\x14.brain.Command.Input\x1a\x43\n\x05Input\x12\x0c\n\x04Name\x18\x01 \x02(\t\x12\x0c\n\x04Type\x18\x02 \x02(\t\x12\x0f\n\x07Tooltip\x18\x03 \x02(\t\x12\r\n\x05Value\x18\x04 \x02(\t\"H\n\x06Target\x12\n\n\x02id\x18\x01 \x01(\t\x12\x12\n\nPluginName\x18\x02 \x02(\t\x12\x10\n\x08Location\x18\x03 \x02(\t\x12\x0c\n\x04Port\x18\x04 \x02(\t\"z\n\x03Job\x12\n\n\x02id\x18\x01 \x01(\t\x12 \n\tJobTarget\x18\x02 \x02(\x0b\x32\r.brain.Target\x12\x0e\n\x06Status\x18\x03 \x02(\t\x12\x11\n\tStartTime\x18\x04 \x02(\x05\x12\"\n\nJobCommand\x18\x05 \x02(\x0b\x32\x0e.brain.Command\"D\n\x06Output\x12\n\n\x02id\x18\x01 \x01(\t\x12\x1d\n\tOutputJob\x18\x02 \x02(\x0b\x32\n.brain.Job\x12\x0f\n\x07\x43ontent\x18\x03 \x02(\t\"e\n\x08JobAudit\x12\n\n\x02id\x18\x01 \x01(\t\x12\x13\n\x0b\x63hange_time\x18\x02 \x02(\x02\x12\x1b\n\x07old_val\x18\x03 \x02(\x0b\x32\n.brain.Job\x12\x1b\n\x07new_val\x18\x04 \x02(\x0b\x32\n.brain.Job\"n\n\x0bTargetAudit\x12\n\n\x02id\x18\x01 \x01(\t\x12\x13\n\x0b\x63hange_time\x18\x02 \x02(\x02\x12\x1e\n\x07old_val\x18\x03 \x02(\x0b\x32\r.brain.Target\x12\x1e\n\x07new_val\x18\x04 \x02(\x0b\x32\r.brain.Target\"e\n\x0bOutputAudit\x12\n\n\x02id\x18\x01 \x01(\t\x12\n\n\x02ts\x18\x02 \x02(\x02\x12\x1e\n\x07old_val\x18\x03 \x02(\x0b\x32\r.brain.Output\x12\x1e\n\x07new_val\x18\x04 \x02(\x0b\x32\r.brain.Output\"\x89\x01\n\x05\x41udit\x12\n\n\x02id\x18\x01 \x01(\t\x12\x1e\n\x03Job\x18\x02 \x01(\x0b\x32\x0f.brain.JobAuditH\x00\x12$\n\x06Target\x18\x03 \x01(\x0b\x32\x12.brain.TargetAuditH\x00\x12$\n\x06Output\x18\x04 \x01(\x0b\x32\x12.brain.OutputAuditH\x00\x42\x08\n\x06record\")\n\x07Targets\x12\x1e\n\x07Targets\x18\x01 \x03(\x0b\x32\r.brain.Target\" \n\x04Jobs\x12\x18\n\x04Jobs\x18\x01 \x03(\x0b\x32\n.brain.Job\",\n\x08\x43ommands\x12 \n\x08\x43ommands\x18\x01 \x03(\x0b\x32\x0e.brain.Command')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_COMMAND_INPUT = _descriptor.Descriptor(
  name='Input',
  full_name='brain.Command.Input',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Name', full_name='brain.Command.Input.Name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Type', full_name='brain.Command.Input.Type', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Tooltip', full_name='brain.Command.Input.Tooltip', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Value', full_name='brain.Command.Input.Value', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=184,
  serialized_end=251,
)

_COMMAND = _descriptor.Descriptor(
  name='Command',
  full_name='brain.Command',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='brain.Command.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='CommandName', full_name='brain.Command.CommandName', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Tooltip', full_name='brain.Command.Tooltip', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Output', full_name='brain.Command.Output', index=3,
      number=4, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Inputs', full_name='brain.Command.Inputs', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='OptionalInputs', full_name='brain.Command.OptionalInputs', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_COMMAND_INPUT, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23,
  serialized_end=251,
)


_TARGET = _descriptor.Descriptor(
  name='Target',
  full_name='brain.Target',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='brain.Target.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='PluginName', full_name='brain.Target.PluginName', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Location', full_name='brain.Target.Location', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Port', full_name='brain.Target.Port', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=253,
  serialized_end=325,
)


_JOB = _descriptor.Descriptor(
  name='Job',
  full_name='brain.Job',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='brain.Job.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='JobTarget', full_name='brain.Job.JobTarget', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Status', full_name='brain.Job.Status', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='StartTime', full_name='brain.Job.StartTime', index=3,
      number=4, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='JobCommand', full_name='brain.Job.JobCommand', index=4,
      number=5, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=327,
  serialized_end=449,
)


_OUTPUT = _descriptor.Descriptor(
  name='Output',
  full_name='brain.Output',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='brain.Output.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='OutputJob', full_name='brain.Output.OutputJob', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Content', full_name='brain.Output.Content', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=451,
  serialized_end=519,
)


_JOBAUDIT = _descriptor.Descriptor(
  name='JobAudit',
  full_name='brain.JobAudit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='brain.JobAudit.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='change_time', full_name='brain.JobAudit.change_time', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='old_val', full_name='brain.JobAudit.old_val', index=2,
      number=3, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='new_val', full_name='brain.JobAudit.new_val', index=3,
      number=4, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=521,
  serialized_end=622,
)


_TARGETAUDIT = _descriptor.Descriptor(
  name='TargetAudit',
  full_name='brain.TargetAudit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='brain.TargetAudit.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='change_time', full_name='brain.TargetAudit.change_time', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='old_val', full_name='brain.TargetAudit.old_val', index=2,
      number=3, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='new_val', full_name='brain.TargetAudit.new_val', index=3,
      number=4, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=624,
  serialized_end=734,
)


_OUTPUTAUDIT = _descriptor.Descriptor(
  name='OutputAudit',
  full_name='brain.OutputAudit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='brain.OutputAudit.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ts', full_name='brain.OutputAudit.ts', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='old_val', full_name='brain.OutputAudit.old_val', index=2,
      number=3, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='new_val', full_name='brain.OutputAudit.new_val', index=3,
      number=4, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=736,
  serialized_end=837,
)


_AUDIT = _descriptor.Descriptor(
  name='Audit',
  full_name='brain.Audit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='brain.Audit.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Job', full_name='brain.Audit.Job', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Target', full_name='brain.Audit.Target', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Output', full_name='brain.Audit.Output', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='record', full_name='brain.Audit.record',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=840,
  serialized_end=977,
)


_TARGETS = _descriptor.Descriptor(
  name='Targets',
  full_name='brain.Targets',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Targets', full_name='brain.Targets.Targets', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=979,
  serialized_end=1020,
)


_JOBS = _descriptor.Descriptor(
  name='Jobs',
  full_name='brain.Jobs',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Jobs', full_name='brain.Jobs.Jobs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1022,
  serialized_end=1054,
)


_COMMANDS = _descriptor.Descriptor(
  name='Commands',
  full_name='brain.Commands',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Commands', full_name='brain.Commands.Commands', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1056,
  serialized_end=1100,
)

_COMMAND_INPUT.containing_type = _COMMAND
_COMMAND.fields_by_name['Inputs'].message_type = _COMMAND_INPUT
_COMMAND.fields_by_name['OptionalInputs'].message_type = _COMMAND_INPUT
_JOB.fields_by_name['JobTarget'].message_type = _TARGET
_JOB.fields_by_name['JobCommand'].message_type = _COMMAND
_OUTPUT.fields_by_name['OutputJob'].message_type = _JOB
_JOBAUDIT.fields_by_name['old_val'].message_type = _JOB
_JOBAUDIT.fields_by_name['new_val'].message_type = _JOB
_TARGETAUDIT.fields_by_name['old_val'].message_type = _TARGET
_TARGETAUDIT.fields_by_name['new_val'].message_type = _TARGET
_OUTPUTAUDIT.fields_by_name['old_val'].message_type = _OUTPUT
_OUTPUTAUDIT.fields_by_name['new_val'].message_type = _OUTPUT
_AUDIT.fields_by_name['Job'].message_type = _JOBAUDIT
_AUDIT.fields_by_name['Target'].message_type = _TARGETAUDIT
_AUDIT.fields_by_name['Output'].message_type = _OUTPUTAUDIT
_AUDIT.oneofs_by_name['record'].fields.append(
  _AUDIT.fields_by_name['Job'])
_AUDIT.fields_by_name['Job'].containing_oneof = _AUDIT.oneofs_by_name['record']
_AUDIT.oneofs_by_name['record'].fields.append(
  _AUDIT.fields_by_name['Target'])
_AUDIT.fields_by_name['Target'].containing_oneof = _AUDIT.oneofs_by_name['record']
_AUDIT.oneofs_by_name['record'].fields.append(
  _AUDIT.fields_by_name['Output'])
_AUDIT.fields_by_name['Output'].containing_oneof = _AUDIT.oneofs_by_name['record']
_TARGETS.fields_by_name['Targets'].message_type = _TARGET
_JOBS.fields_by_name['Jobs'].message_type = _JOB
_COMMANDS.fields_by_name['Commands'].message_type = _COMMAND
DESCRIPTOR.message_types_by_name['Command'] = _COMMAND
DESCRIPTOR.message_types_by_name['Target'] = _TARGET
DESCRIPTOR.message_types_by_name['Job'] = _JOB
DESCRIPTOR.message_types_by_name['Output'] = _OUTPUT
DESCRIPTOR.message_types_by_name['JobAudit'] = _JOBAUDIT
DESCRIPTOR.message_types_by_name['TargetAudit'] = _TARGETAUDIT
DESCRIPTOR.message_types_by_name['OutputAudit'] = _OUTPUTAUDIT
DESCRIPTOR.message_types_by_name['Audit'] = _AUDIT
DESCRIPTOR.message_types_by_name['Targets'] = _TARGETS
DESCRIPTOR.message_types_by_name['Jobs'] = _JOBS
DESCRIPTOR.message_types_by_name['Commands'] = _COMMANDS

Command = _reflection.GeneratedProtocolMessageType('Command', (_message.Message,), dict(

  Input = _reflection.GeneratedProtocolMessageType('Input', (_message.Message,), dict(
    DESCRIPTOR = _COMMAND_INPUT,
    __module__ = 'brain_pb2'
    # @@protoc_insertion_point(class_scope:brain.Command.Input)
    ))
  ,
  DESCRIPTOR = _COMMAND,
  __module__ = 'brain_pb2'
  # @@protoc_insertion_point(class_scope:brain.Command)
  ))
_sym_db.RegisterMessage(Command)
_sym_db.RegisterMessage(Command.Input)

Target = _reflection.GeneratedProtocolMessageType('Target', (_message.Message,), dict(
  DESCRIPTOR = _TARGET,
  __module__ = 'brain_pb2'
  # @@protoc_insertion_point(class_scope:brain.Target)
  ))
_sym_db.RegisterMessage(Target)

Job = _reflection.GeneratedProtocolMessageType('Job', (_message.Message,), dict(
  DESCRIPTOR = _JOB,
  __module__ = 'brain_pb2'
  # @@protoc_insertion_point(class_scope:brain.Job)
  ))
_sym_db.RegisterMessage(Job)

Output = _reflection.GeneratedProtocolMessageType('Output', (_message.Message,), dict(
  DESCRIPTOR = _OUTPUT,
  __module__ = 'brain_pb2'
  # @@protoc_insertion_point(class_scope:brain.Output)
  ))
_sym_db.RegisterMessage(Output)

JobAudit = _reflection.GeneratedProtocolMessageType('JobAudit', (_message.Message,), dict(
  DESCRIPTOR = _JOBAUDIT,
  __module__ = 'brain_pb2'
  # @@protoc_insertion_point(class_scope:brain.JobAudit)
  ))
_sym_db.RegisterMessage(JobAudit)

TargetAudit = _reflection.GeneratedProtocolMessageType('TargetAudit', (_message.Message,), dict(
  DESCRIPTOR = _TARGETAUDIT,
  __module__ = 'brain_pb2'
  # @@protoc_insertion_point(class_scope:brain.TargetAudit)
  ))
_sym_db.RegisterMessage(TargetAudit)

OutputAudit = _reflection.GeneratedProtocolMessageType('OutputAudit', (_message.Message,), dict(
  DESCRIPTOR = _OUTPUTAUDIT,
  __module__ = 'brain_pb2'
  # @@protoc_insertion_point(class_scope:brain.OutputAudit)
  ))
_sym_db.RegisterMessage(OutputAudit)

Audit = _reflection.GeneratedProtocolMessageType('Audit', (_message.Message,), dict(
  DESCRIPTOR = _AUDIT,
  __module__ = 'brain_pb2'
  # @@protoc_insertion_point(class_scope:brain.Audit)
  ))
_sym_db.RegisterMessage(Audit)

Targets = _reflection.GeneratedProtocolMessageType('Targets', (_message.Message,), dict(
  DESCRIPTOR = _TARGETS,
  __module__ = 'brain_pb2'
  # @@protoc_insertion_point(class_scope:brain.Targets)
  ))
_sym_db.RegisterMessage(Targets)

Jobs = _reflection.GeneratedProtocolMessageType('Jobs', (_message.Message,), dict(
  DESCRIPTOR = _JOBS,
  __module__ = 'brain_pb2'
  # @@protoc_insertion_point(class_scope:brain.Jobs)
  ))
_sym_db.RegisterMessage(Jobs)

Commands = _reflection.GeneratedProtocolMessageType('Commands', (_message.Message,), dict(
  DESCRIPTOR = _COMMANDS,
  __module__ = 'brain_pb2'
  # @@protoc_insertion_point(class_scope:brain.Commands)
  ))
_sym_db.RegisterMessage(Commands)


# @@protoc_insertion_point(module_scope)





from google.protobuf.message import EncodeError
import dict_to_protobuf #this lib allows extra keys
dict_to_protobuf.l.setLevel("ERROR")

def verify(value, msg):
    """
    C-style validator

    Keyword arguments:
    value -- dictionary to validate (required)
    msg -- the protobuf schema to validate against (required)

    Returns:
        True: If valid input
        False: If invalid input
    """
    result = True
    dict_to_protobuf.dict_to_protobuf(value, msg)
    try:
        msg.SerializeToString()
    except EncodeError:
        result = False
    return result

def strip(value, msg):
    """
    Strips all non-essential keys from the value dictionary
    given the message format protobuf

    raises ValueError exception if value does not have all required keys

    :param value: <dict> with arbitrary keys
    :param msg: <protobuf> with a defined schema
    :return: NEW <dict> with keys defined by msg, omits any other key
    """
    dict_to_protobuf.dict_to_protobuf(value, msg)
    try:
        msg.SerializeToString()  #raise error for insufficient input
    except EncodeError as encode_error:
        raise ValueError(str(encode_error))
    output = dict_to_protobuf.protobuf_to_dict(msg)
    return output
